using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application.TicketGroups.Queries.GetTicketGroup;

public class GetTicketGroupQueryHandler :
    IRequestHandler<GetTicketGroupQuery, TicketGroupDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly CurrencyConverterService _currencyConverter;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly SessionTimeZoneService _sessionTimeZoneService;

    private readonly object _lock = new();

    public GetTicketGroupQueryHandler(UnitOfWork unitOfWork,
        IMapper mapper, CurrencyConverterService currencyConverterService,
        SessionCurrencyService sessionCurrencyService,
        SessionTimeZoneService sessionTimeZoneService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _currencyConverter = currencyConverterService;
        _sessionCurrencyService = sessionCurrencyService;
        _sessionTimeZoneService = sessionTimeZoneService;
    }

    public async Task<TicketGroupDto> Handle(
        GetTicketGroupQuery request,
        CancellationToken cancellationToken)
    {
        var ticketGroup = await _unitOfWork.TicketGroupRepository.GetOneAsync(
            e => e.Guid == request.Guid, e => e.Tickets,
            cancellationToken);

        if (ticketGroup == null)
        {
            throw new NotFoundException();
        }


        // Hydrate

        var vehicleEnrollmentIds =
            ticketGroup.Tickets.Select(t => t.VehicleEnrollmentId);
        var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
            .GetPageAsync(
                ve => vehicleEnrollmentIds.Contains(ve.Id),
                ve => ve.Route.RouteAddresses,
                1, vehicleEnrollmentIds.Count(), cancellationToken))
            .Items;

        var routeAddressIds = vehicleEnrollments
            .SelectMany(ve => ve.Route.RouteAddresses)
            .Select(ra => ra.Id);
        var routeAddressDetails = (await _unitOfWork.RouteAddressDetailRepository
            .GetPageAsync(
                rad => routeAddressIds.Contains(rad.RouteAddressId),
                1, routeAddressIds.Count(), cancellationToken))
            .Items;

        var addressIds = vehicleEnrollments
            .SelectMany(ve => ve.Route.RouteAddresses)
            .Select(ra => ra.AddressId);
        var addresses = (await _unitOfWork.AddressRepository
            .GetPageAsync(
                a => addressIds.Contains(a.Id),
                a => a.City.Region.Country,
                1, addressIds.Count(), cancellationToken))
            .Items;

        var vehicleIds = vehicleEnrollments
            .Select(ve => ve.VehicleId);
        var vehicles = (await _unitOfWork.VehicleRepository
            .GetPageAsync(
                v => vehicleIds.Contains(v.Id),
                v => v.Company,
                1, vehicleIds.Count(), cancellationToken))
            .Items;

        var account = await _unitOfWork.AccountRepository.GetOneAsync(
            a => a.Id == ticketGroup.AccountId, cancellationToken);

        if (ticketGroup.AccountId != null)
        {
            ticketGroup.Account = account;
        }

        foreach (var ve in vehicleEnrollments)
        {
            ve.Vehicle = vehicles.Single(v => v.Id == ve.VehicleId);

            foreach (var ra in ve.Route.RouteAddresses)
            {
                ra.Address = addresses.Single(a => a.Id == ra.AddressId);
                ra.Details = routeAddressDetails
                    .Where(rad => rad.RouteAddressId == ra.Id)
                    .ToArray();
            }
        }


        // TODO: Replace with AutoMapper resolvers
        // Convert currency and apply session time zone

        var convertTasks = new List<Task>();

        foreach (var t in ticketGroup.Tickets)
        {
            convertTasks.Add(Task.Factory.StartNew(() =>
            {
                t.VehicleEnrollment.DepartureTime =
                    TimeZoneInfo.ConvertTime(t.VehicleEnrollment.DepartureTime,
                        _sessionTimeZoneService.TimeZone);
            }));

            if (_sessionCurrencyService.Currency.Equals(Currency.Default))
            {
                break;
            }

            convertTasks.Add(Task.Factory.StartNew(() =>
            {
                lock (_lock)
                {
                    var convertedCost = _currencyConverter.ConvertAsync(t.Cost,
                        t.Currency, _sessionCurrencyService.Currency,
                        cancellationToken)
                    .Result;
                    
                    t.Cost = _sessionCurrencyService
                        .Currency.Round(convertedCost);
                }
            }));

            foreach (var rad in t.VehicleEnrollment.RouteAddressDetails)
            {
                convertTasks.Add(Task.Factory.StartNew(() =>
                {
                    lock (_lock)
                    {
                        var convertedCost = _currencyConverter.ConvertAsync(
                           rad.CostToNextAddress, t.VehicleEnrollment.Currency,
                           _sessionCurrencyService.Currency, cancellationToken)
                        .Result;

                        rad.CostToNextAddress = _sessionCurrencyService
                            .Currency.Round(convertedCost);
                    }
                }));
            }
        }

        Task.WaitAll(convertTasks);

        foreach (var t in ticketGroup.Tickets)
        {
            if (_sessionCurrencyService.Currency.Equals(Currency.Default))
            {
                break;
            }

            t.Currency = _sessionCurrencyService.Currency;
            t.VehicleEnrollment.Currency = _sessionCurrencyService.Currency;
        }


        _unitOfWork.Dispose();

        var dto = _mapper.Map<TicketGroupDto>(ticketGroup);

        dto.Currency = _sessionCurrencyService.Currency.Name;

        return dto;
    }
}
