using MediatR;
using cuqmbr.TravelGuide.Application.Common.Persistence;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.Extensions;
using cuqmbr.TravelGuide.Application.Common.Services;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application.TicketGroups.Queries.GetTicketGroupsPage;

public class GetTicketGroupsPageQueryHandler :
    IRequestHandler<GetTicketGroupsPageQuery, PaginatedList<TicketGroupDto>>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly CurrencyConverterService _currencyConverter;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly SessionTimeZoneService _sessionTimeZoneService;

    private readonly object _lock = new();

    public GetTicketGroupsPageQueryHandler(UnitOfWork unitOfWork,
        IMapper mapper, CurrencyConverterService currencyConverterService,
        SessionCurrencyService sessionCurrencyService,
        SessionTimeZoneService sessionTimeZoneService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _currencyConverter = currencyConverterService;
        _sessionCurrencyService = sessionCurrencyService;
        _sessionTimeZoneService = sessionTimeZoneService;
    }

    public async Task<PaginatedList<TicketGroupDto>> Handle(
        GetTicketGroupsPageQuery request,
        CancellationToken cancellationToken)
    {
        var paginatedList = await _unitOfWork.TicketGroupRepository.GetPageAsync(
            e =>
                (e.PassangerFirstName.ToLower().Contains(request.Search.ToLower()) ||
                e.PassangerLastName.ToLower().Contains(request.Search.ToLower()) ||
                e.PassangerPatronymic.ToLower().Contains(request.Search.ToLower()) ||
                (e.PassangerEmail != null ?
                    e.PassangerEmail.ToLower().Contains(request.Search.ToLower()) :
                    false)) &&
                (request.PassangerSex != null
                    ? request.PassangerSex.Contains(e.PassangerSex)
                    : true) &&
                (request.PassangerBirthDateGreaterThanOrEqualTo != null
                    ? e.PassangerBirthDate >= request.PassangerBirthDateGreaterThanOrEqualTo
                    : true) &&
                (request.PassangerBirthDateLessThanOrEqualTo != null
                    ? e.PassangerBirthDate <= request.PassangerBirthDateLessThanOrEqualTo
                    : true) &&
                (request.PurchaseTimeGreaterThanOrEqualTo != null
                    ? e.PurchaseTime >= request.PurchaseTimeGreaterThanOrEqualTo
                    : true) &&
                (request.PurchaseTimeLessThanOrEqualTo != null
                    ? e.PurchaseTime <= request.PurchaseTimeLessThanOrEqualTo
                    : true) &&
                (request.PassangerSex != null
                    ? request.PassangerSex.Contains(e.PassangerSex)
                    : true) &&
                (request.Statuses != null
                    ? request.Statuses.Contains(e.Status)
                    : true) &&
                (request.VehicleTypes != null
                    ? e.Tickets
                        .Select(t => t.VehicleEnrollment.Vehicle.VehicleType)
                        .Any(vt => request.VehicleTypes.Contains(vt)) 
                    : true) &&
                (request.TravelTimeGreaterThanOrEqualTo != null
                    ? e.TravelTime >= request.TravelTimeGreaterThanOrEqualTo
                    : true) &&
                (request.TravelTimeLessThanOrEqualTo != null
                    ? e.TravelTime <= request.TravelTimeLessThanOrEqualTo
                    : true) &&
                (request.AccountGuid != null
                    ? e.Account.Guid == request.AccountGuid
                    : true),
            e => e.Tickets,
            request.PageNumber, request.PageSize, cancellationToken);

        var ticketGroups = paginatedList.Items;


        // Hydrate

        var vehicleEnrollmentIds =
            ticketGroups.SelectMany(tg => tg.Tickets)
            .Select(t => t.VehicleEnrollmentId);
        var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
            .GetPageAsync(
                ve => vehicleEnrollmentIds.Contains(ve.Id),
                ve => ve.Route.RouteAddresses,
                1, vehicleEnrollmentIds.Count(), cancellationToken))
            .Items;

        var routeAddressIds = vehicleEnrollments
            .SelectMany(ve => ve.Route.RouteAddresses)
            .Select(ra => ra.Id);
        var routeAddressDetails = (await _unitOfWork.RouteAddressDetailRepository
            .GetPageAsync(
                rad => routeAddressIds.Contains(rad.RouteAddressId),
                1, routeAddressIds.Count(), cancellationToken))
            .Items;

        var addressIds = vehicleEnrollments
            .SelectMany(ve => ve.Route.RouteAddresses)
            .Select(ra => ra.AddressId);
        var addresses = (await _unitOfWork.AddressRepository
            .GetPageAsync(
                a => addressIds.Contains(a.Id),
                a => a.City.Region.Country,
                1, addressIds.Count(), cancellationToken))
            .Items;

        var vehicleIds = vehicleEnrollments
            .Select(ve => ve.VehicleId);
        var vehicles = (await _unitOfWork.VehicleRepository
            .GetPageAsync(
                v => vehicleIds.Contains(v.Id),
                v => v.Company,
                1, vehicleIds.Count(), cancellationToken))
            .Items;

        var accountIds =
            ticketGroups.Select(tg => tg.AccountId);
        var accounts = (await _unitOfWork.AccountRepository
            .GetPageAsync(
                a => accountIds.Contains(a.Id),
                1, accountIds.Count(), cancellationToken))
            .Items;

        foreach (var tg in ticketGroups)
        {
            if (tg.AccountId != null)
            {
                tg.Account = accounts.Single(a => a.Id == tg.AccountId);
            }
        }

        foreach (var ve in vehicleEnrollments)
        {
            ve.Vehicle = vehicles.Single(v => v.Id == ve.VehicleId);

            foreach (var ra in ve.Route.RouteAddresses)
            {
                ra.Address = addresses.Single(a => a.Id == ra.AddressId);
                ra.Details = routeAddressDetails
                    .Where(rad => rad.RouteAddressId == ra.Id)
                    .ToArray();
            }
        }


        // TODO: Replace with AutoMapper resolvers.
        // Convert currency and apply session time zone.

        var convertTasks = new List<Task>();
        var processedRouteAddressDetailIds = new HashSet<long>();

        foreach (var t in ticketGroups.SelectMany(tg => tg.Tickets))
        {
            convertTasks.Add(Task.Factory.StartNew(() =>
            {
                t.VehicleEnrollment.DepartureTime =
                    TimeZoneInfo.ConvertTime(t.VehicleEnrollment.DepartureTime,
                        _sessionTimeZoneService.TimeZone);
            }));

            if (_sessionCurrencyService.Currency.Equals(Currency.Default))
            {
                break;
            }

            convertTasks.Add(Task.Factory.StartNew(() =>
            {
                lock (_lock)
                {
                    var convertedCost = _currencyConverter.ConvertAsync(t.Cost,
                        t.Currency, _sessionCurrencyService.Currency,
                        t.TicketGroup.PurchaseTime, cancellationToken)
                    .Result;

                    t.Cost = _sessionCurrencyService
                        .Currency.Round(convertedCost);
                }
            }));

            foreach (var rad in t.VehicleEnrollment.RouteAddressDetails)
            {
                convertTasks.Add(Task.Factory.StartNew(() =>
                {
                    lock (_lock)
                    {
                        if (processedRouteAddressDetailIds.Contains(rad.Id))
                        {
                            return;
                        }

                        var convertedCost = _currencyConverter.ConvertAsync(
                           rad.CostToNextAddress, t.VehicleEnrollment.Currency,
                           _sessionCurrencyService.Currency,
                           t.TicketGroup.PurchaseTime, cancellationToken)
                        .Result;

                        rad.CostToNextAddress = _sessionCurrencyService
                            .Currency.Round(convertedCost);

                        processedRouteAddressDetailIds.Add(rad.Id);
                    }
                }));
            }
        }

        Task.WaitAll(convertTasks);

        foreach (var t in ticketGroups.SelectMany(tg => tg.Tickets))
        {
            if (_sessionCurrencyService.Currency.Equals(Currency.Default))
            {
                break;
            }

            t.Currency = _sessionCurrencyService.Currency;
            t.VehicleEnrollment.Currency = _sessionCurrencyService.Currency;
        }


        var mappedItems =
            _mapper.Map<IEnumerable<TicketGroupDto>>(ticketGroups);

        foreach (var item in mappedItems)
        {
            item.Currency = _sessionCurrencyService.Currency.Name;
        }

        mappedItems = QueryableExtension<TicketGroupDto>
            .ApplySort(mappedItems.AsQueryable(), request.Sort);

        _unitOfWork.Dispose();

        return new PaginatedList<TicketGroupDto>(
            mappedItems.ToList(),
            paginatedList.TotalCount, request.PageNumber,
            request.PageSize);
    }
}
