using Microsoft.AspNetCore.Mvc;
using Swashbuckle.AspNetCore.Annotations;
using cuqmbr.TravelGuide.Domain.Enums;
using cuqmbr.TravelGuide.Application.TicketGroups;
using cuqmbr.TravelGuide.Application.TicketGroups.Commands.AddTicketGroup;
using cuqmbr.TravelGuide.Application.TicketGroups.Queries.GetTicketGroup;
using cuqmbr.TravelGuide.Application.TicketGroups.Queries.GetTicketGroupsPage;
using cuqmbr.TravelGuide.Application.TicketGroups.ViewModels;
using cuqmbr.TravelGuide.Application.TicketGroups.Models;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.ViewModels;

namespace cuqmbr.TravelGuide.HttpApi.Controllers;

[Route("ticketGroups")]
public class TicketGroupsController : ControllerBase
{
    [HttpPost]
    [SwaggerOperation("Add a ticket group")]
    [SwaggerResponse(
        StatusCodes.Status201Created, "Object successfuly created",
        typeof(TicketGroupDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Object already exists",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Parent object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<ActionResult<TicketGroupDto>> Add(
        [FromBody] AddTicketGroupViewModel viewModel,
        CancellationToken cancellationToken)
    {
        return StatusCode(
            StatusCodes.Status201Created,
            await Mediator.Send(
                new AddTicketGroupCommand()
                {
                    PassangerFirstName = viewModel.PassangerFirstName,
                    PassangerLastName = viewModel.PassangerLastName,
                    PassangerPatronymic = viewModel.PassangerPatronymic,
                    PassangerSex = Sex.FromName(viewModel.PassangerSex),
                    PassangerBirthDate = viewModel.PassangerBirthDate,
                    PurchaseTime = viewModel.PurchaseTime,
                    Status = TicketStatus.FromName(viewModel.Status),
                    Tickets = viewModel.Tickets.Select(e =>
                        new TicketModel()
                        {
                            DepartureRouteAddressGuid = 
                                e.DepartureRouteAddressUuid,
                            ArrivalRouteAddressGuid =
                                e.ArrivalRouteAddressUuid,
                            Order = e.Order,
                            VehicleEnrollmentGuid =
                                e.VehicleEnrollmentUuid,
                        })
                        .ToArray()
                },
                cancellationToken));
    }

    [HttpGet]
    [SwaggerOperation("Get a list of all ticket groups")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful",
        typeof(PaginatedList<TicketGroupDto>))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<PaginatedList<TicketGroupDto>> GetPage(
        [FromQuery] PageQuery pageQuery, [FromQuery] SearchQuery searchQuery,
        [FromQuery] SortQuery sortQuery, 
        [FromQuery] GetTicketGroupsPageFilterViewModel filterQuery,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(
            new GetTicketGroupsPageQuery()
            {
                PageNumber = pageQuery.PageNumber,
                PageSize = pageQuery.PageSize,
                Search = searchQuery.Search,
                Sort = sortQuery.Sort,
                PassangerSex = filterQuery.PassangerSex?
                    .Select(s => Sex.FromName(s)).ToHashSet(),
                PassangerBirthDateGreaterThanOrEqualTo = 
                    filterQuery.PassangerBirthDateGreaterThanOrEqualTo,
                PassangerBirthDateLessThanOrEqualTo = 
                    filterQuery.PassangerBirthDateLessThanOrEqualTo,
                PurchaseTimeGreaterThanOrEqualTo = 
                    filterQuery.PurchaseTimeGreaterThanOrEqualTo,
                PurchaseTimeLessThanOrEqualTo = 
                    filterQuery.PurchaseTimeLessThanOrEqualTo,
                Statuses = filterQuery.Statuses?
                    .Select(s => TicketStatus.FromName(s)).ToHashSet(),
                VehicleTypes = filterQuery.VehicleTypes?
                    .Select(vt => VehicleType.FromName(vt)).ToHashSet(),
                TravelTimeGreaterThanOrEqualTo =
                    filterQuery.TravelTimeGreaterThanOrEqualTo,
                TravelTimeLessThanOrEqualTo = 
                    filterQuery.TravelTimeLessThanOrEqualTo,
                AccountGuid = filterQuery.AccountUuid
            },
            cancellationToken);
    }

    [HttpGet("{uuid:guid}")]
    [SwaggerOperation("Get a ticket group by uuid")]
    [SwaggerResponse(
        StatusCodes.Status200OK, "Request successful", typeof(TicketGroupDto))]
    [SwaggerResponse(
        StatusCodes.Status400BadRequest, "Input data validation error",
        typeof(HttpValidationProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status401Unauthorized, "Unauthorized to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status403Forbidden,
        "Not enough privileges to perform an action",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status404NotFound, "Object not found",
        typeof(ProblemDetails))]
    [SwaggerResponse(
        StatusCodes.Status500InternalServerError, "Internal server error",
        typeof(ProblemDetails))]
    public async Task<TicketGroupDto> Get(
        [FromRoute] Guid uuid,
        CancellationToken cancellationToken)
    {
        return await Mediator.Send(new GetTicketGroupQuery() { Guid = uuid },
            cancellationToken);
    }
}
